from traitlets import Bool, List, Unicode, observe
import numpy as np
import os
import regions

from glue.core.message import DataCollectionAddMessage, DataCollectionDeleteMessage
from glue_jupyter.common.toolbar_vuetify import read_icon

from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.events import LinkUpdatedMessage, ChangeRefDataMessage
from jdaviz.core.marks import FootprintOverlay
from jdaviz.core.region_translators import regions2roi
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, ViewerSelectMixin,
                                        EditableSelectPluginComponent, SelectPluginComponent,
                                        FileImportSelectPluginComponent, HasFileImportSelect)
from jdaviz.core.tools import ICON_DIR
from jdaviz.core.user_api import PluginUserApi

from jdaviz.configs.imviz.plugins.footprints import preset_regions


__all__ = ['Footprints']


@tray_registry('imviz-footprints', label="Footprints")
class Footprints(PluginTemplateMixin, ViewerSelectMixin, HasFileImportSelect):
    """
    See the :ref:`Footprints Plugin Documentation <imviz-footprints>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``overlay`` (:class:`~jdaviz.core.template_mixin.EditableSelectPluginComponent`): the
        currently active overlay (all other traitlets control this overlay instance)
    * :meth:`rename_overlay`
        rename any overlay
    * :meth:`add_overlay`
        add a new overlay instance (and select as active)
    * :meth:`remove_overlay`
        remove any overlay
    * ``viewer`` (:class:`~jdaviz.core.template_mixin.ViewerSelect`):
        viewer(s) to show the current overlay
    * ``visible``
        whether the currently selected overlay should be visible in the selected viewers
    * ``color``
        color of the currently selected overlay
    * ``fill_opacity``
        opacity of the filled region of the currently selected overlay
    * ``preset_obs`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
        selected observatories to filter ``preset`` choices.
    * ``preset`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
        selected overlay preset
    * :meth:`import_region`
    * :meth:`center_on_viewer`
    * ``ra``
        central right ascension for the footprint overlay
    * ``dec``
        central declination (in degrees) for the footprint overlay
    * ``pa``
        position angle (in degrees) measured from North to central vertical axis in North to East
        direction.
    * ``v2_offset``
        Additional V2 offset in telescope coordinates to apply to instrument center, as from a
        dither pattern.
    * ``v3_offset``
        Additional V3 offset in telescope coordinates to apply to instrument center, as from a
        dither pattern.
    * :meth:`overlay_regions`
    """
    template_file = __file__, "footprints.vue"
    uses_active_status = Bool(True).tag(sync=True)

    is_pixel_linked = Bool(True).tag(sync=True)  # plugin disabled if linked by pixel (default)

    # OVERLAY LABEL
    overlay_mode = Unicode().tag(sync=True)
    overlay_edit_value = Unicode().tag(sync=True)
    overlay_items = List().tag(sync=True)
    overlay_selected = Unicode().tag(sync=True)

    # STYLING OPTIONS
    # viewer (via mixin)
    visible = Bool(True).tag(sync=True)
    color = Unicode('#c75109').tag(sync=True)  # "active" orange
    fill_opacity = FloatHandleEmpty(0.1).tag(sync=True)

    # PRESET OVERLAYS AND OPTIONS
    has_pysiaf = Bool(preset_regions._has_pysiaf).tag(sync=True)
    preset_obs_items = List().tag(sync=True)
    preset_obs_selected = Unicode().tag(sync=True)
    preset_items = List().tag(sync=True)
    preset_selected = Unicode().tag(sync=True)

    ra = FloatHandleEmpty().tag(sync=True)
    dec = FloatHandleEmpty().tag(sync=True)
    pa = FloatHandleEmpty().tag(sync=True)
    v2_offset = FloatHandleEmpty().tag(sync=True)
    v3_offset = FloatHandleEmpty().tag(sync=True)
    # TODO: dithering/mosaic options?

    def __init__(self, *args, **kwargs):
        self._ignore_traitlet_change = False
        self._overlays = {}

        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Show instrument footprints as overlays on image viewers.'

        self.viewer.multiselect = True  # multiselect always enabled
        # require a viewer's reference data to have WCS so that footprints can be mapped to sky
        self.viewer.add_filter('reference_has_wcs')

        self.overlay = EditableSelectPluginComponent(self,
                                                     name='overlay',
                                                     mode='overlay_mode',
                                                     edit_value='overlay_edit_value',
                                                     items='overlay_items',
                                                     selected='overlay_selected',
                                                     manual_options=['default'],
                                                     on_rename=self._on_overlay_rename,
                                                     on_remove=self._on_overlay_remove)

        if self.has_pysiaf:
            obs_icons = {'JWST': read_icon(os.path.join(ICON_DIR, 'jwst_solid.svg'), 'svg+xml'),
                         'Roman': read_icon(os.path.join(ICON_DIR, 'roman_solid.svg'), 'svg+xml')}
            preset_options = [{'label': display_name,
                               'siaf_name': siaf_name,
                               'observatory': observatory,
                               'icon': obs_icons.get(observatory, None)}
                              for observatory, instruments in preset_regions._instruments.items()
                              for display_name, siaf_name in instruments.items()]
            preset_obs_options = ['Any'] + [{'label': obs, 'icon': obs_icons.get(obs)}
                                            for obs in preset_regions._instruments.keys()]
        else:
            preset_options = ['None']
            preset_obs_options = []

        self.preset_obs = SelectPluginComponent(self,
                                                items='preset_obs_items',
                                                selected='preset_obs_selected',
                                                manual_options=preset_obs_options)

        self.preset = FileImportSelectPluginComponent(self,
                                                      items='preset_items',
                                                      selected='preset_selected',
                                                      manual_options=preset_options,
                                                      apply_filters_to_manual_options=True,
                                                      server_is_remote=self.app.state.settings.get('server_is_remote', False))  # noqa

        # set the custom file parser for importing catalogs
        self.preset._file_parser = self._file_parser

        # disable if pixel-linked AND only a single item in the data collection
        self.hub.subscribe(self, LinkUpdatedMessage, handler=self._on_link_type_updated)
        self.hub.subscribe(self, DataCollectionAddMessage, handler=self._on_link_type_updated)
        self.hub.subscribe(self, DataCollectionDeleteMessage, handler=self._on_link_type_updated)
        self.hub.subscribe(self, ChangeRefDataMessage, handler=self._on_rotation)  # noqa
        self._on_link_type_updated()

    @property
    def user_api(self):
        return PluginUserApi(self, expose=('overlay',
                                           'rename_overlay', 'add_overlay', 'remove_overlay',
                                           'viewer', 'visible', 'color', 'fill_opacity',
                                           'preset_obs', 'preset', 'import_region',
                                           'center_on_viewer', 'ra', 'dec', 'pa',
                                           'v2_offset', 'v3_offset',
                                           'overlay_regions'))

    def _get_marks(self, viewer, overlay=None):
        matches = [mark for mark in viewer.figure.marks
                   if (isinstance(mark, FootprintOverlay) and
                       (mark.overlay == overlay or overlay is None))]
        return matches

    @property
    def marks(self):
        # {overlay: {viewer: [list_of_marks]}}
        return {overlay: {viewer_id: self._get_marks(viewer, overlay)
                          for viewer_id, viewer in self.app._viewer_store.items()
                          if hasattr(viewer, 'figure')}
                for overlay in self.overlay.choices}

    @staticmethod
    def _file_parser(path):
        def _ensure_sky(region):
            if isinstance(region, regions.Regions):
                return np.all([_ensure_sky(reg) for reg in region.regions])
            return hasattr(region, 'to_pixel')

        if isinstance(path, (regions.Region, regions.Regions)):
            if not _ensure_sky(path):
                return 'Region is not a SkyRegion', {}
            from_file_string = f'API: {path.__class__.__name__} object'
            return '', {from_file_string: path}

        try:
            region = regions.Regions.read(path)
        except Exception:
            return 'Could not parse region from file', {}
        if not _ensure_sky(region):
            return 'Region is not a SkyRegion', {}
        return '', {path: region}

    def _on_link_type_updated(self, msg=None):
        self.is_pixel_linked = (getattr(self.app, '_align_by', None) == 'pixels' and
                                len(self.app.data_collection) > 1)
        # toggle visibility as necessary
        self._on_is_active_changed()

        # When footprint(s) are added via API before WCS link. Overlays visibility & is_active
        # can be True, but only last footprint will display. This ensures all footprints display
        if not self.is_pixel_linked:
            for choice in self.overlay.choices:
                # trigger the update without actually changing the user-selection
                self._change_overlay(overlay_selected=choice, center_the_overlay=False)

    def vue_link_by_wcs(self, *args):
        # call other plugin so that other options (wcs_fast_approximation, wcs_use_fallback)
        # are retained.  Remove this method if support for plotting footprints
        # when pixel-linked is reintroduced.
        op = self.app._jdaviz_helper.plugins['Orientation']
        if op._obj.need_clear_astrowidget_markers or op._obj.need_clear_subsets:
            op.open_in_tray()
        else:
            op.align_by = 'WCS'

    def _ensure_first_overlay(self):
        if not len(self._overlays):
            # create the first default overlay
            self._change_overlay()
            # update the marks
            self._preset_args_changed()

    def rename_overlay(self, old_lbl, new_lbl):
        """
        Rename an existing overlay instance

        Parameters
        ----------
        old_lbl : string
            current label of the overlay
        new_lbl : string
            new label of the overlay
        """
        # NOTE: the overlay will call _on_overlay_rename after updating
        self.overlay.rename_choice(old_lbl, new_lbl)

    def add_overlay(self, lbl):
        """
        Add a new overlay instance, and set it as selected.  Once selected, the traitlets
        will then control the options of the new overlay.

        Parameters
        ----------
        lbl : string
            label of the overlay
        """
        # TODO: ability to pass options which would avoid updating the marks until all are set,
        # probably by setattr(self.user_api, k, v) (and checks in advance that all are valid?)
        self.overlay.add_choice(lbl, set_as_selected=True)

    def remove_overlay(self, lbl):
        """
        Remove a overlay instance.  If selected, the selected overlay will default to the
        first entry in the list.

        Parameters
        ----------
        lbl : string
            label of the overlay
        """
        # NOTE: the overlay will call _on_overlay_remove after updating
        self.overlay.remove_choice(lbl)

    def _on_overlay_rename(self, old_lbl, new_lbl):
        # this is triggered when the plugin overlay detects a change to the overlay name
        self._overlays[new_lbl] = self._overlays.pop(old_lbl, {})

        # change the reference on any marks entries for this overlay (in any viewer)
        for viewer_id, viewer in self.app._viewer_store.items():
            if not hasattr(viewer, 'figure'):  # pragma: nocover
                # should only be table viewers in mosviz, but will leave this in case we
                # ever enable the plugin for the image-viewer in mosviz
                continue
            for mark in self._get_marks(viewer, old_lbl):
                mark._overlay = new_lbl

    def _on_overlay_remove(self, lbl):
        _ = self._overlays.pop(lbl, {})

        # remove any marks objects (corresponding to this overlay) from all figures
        for viewer_id, viewer in self.app._viewer_store.items():
            if not hasattr(viewer, 'figure'):  # pragma: nocover
                # should only be table viewers in mosviz, but will leave this in case we
                # ever enable the plugin for the image-viewer in mosviz
                continue
            viewer.figure.marks = [m for m in viewer.figure.marks
                                   if getattr(m, 'overlay', None) != lbl]

    @observe('is_active', 'viewer_items')
    # NOTE: intentionally not using skip_if_no_updates_since_last_active since this only controls
    # visibility of overlay (and creating the first instance)
    def _on_is_active_changed(self, *args):
        if not hasattr(self, 'overlay'):  # pragma: nocover
            # plugin/traitlet startup
            return
        if len(self.disabled_msg):
            # no pysiaf, we don't want to try updating overlays
            return

        if not len(self.viewer.choices):
            # no valid viewers to show footprints
            # TODO: if a viewer becomes valid, we want to call the block below
            return

        if self.is_active:
            self._ensure_first_overlay()

        for overlay, viewer_marks in self.marks.items():
            for viewer_id, marks in viewer_marks.items():
                visible = self._mark_visible(viewer_id, overlay)
                for mark in marks:
                    mark.visible = visible

    def center_on_viewer(self, viewer_ref=None):
        """
        Center the values of RA and DEC based on the current zoom-limits of a viewer.

        Parameters
        ----------
        viewer_ref : string, optional
            Reference of the viewer to center, will default to the first selected viewer
        """
        if viewer_ref is None:
            if not len(self.viewer.selected):  # pragma: nocover
                raise ValueError("no viewers selected, provide viewer reference")
            viewer_ref = self.viewer.selected[0]
        viewer = self.app.get_viewer(viewer_ref)
        center_coord = viewer._get_center_skycoord()
        self._ignore_traitlet_change = True
        self.ra = center_coord.ra.to_value('deg')
        self.dec = center_coord.dec.to_value('deg')
        self._ignore_traitlet_change = False
        self._preset_args_changed()  # process ra/dec simultaneously

    def vue_center_on_viewer(self, viewer_ref):
        self.center_on_viewer(viewer_ref)

    def _on_rotation(self, msg={}):
        for overlay_selected in self._overlays:
            self._change_overlay(overlay_selected=overlay_selected)

    @observe('overlay_selected')
    def _change_overlay(self, *args, overlay_selected=None, center_the_overlay=True):
        if not hasattr(self, 'overlay'):  # pragma: nocover
            # plugin/traitlet startup
            return
        if self.overlay_selected == '':
            # no overlay selected (this can happen when removing all overlays)
            return

        overlay_selected = overlay_selected if overlay_selected is not None else self.overlay_selected  # noqa

        if overlay_selected not in self._overlays:
            # _on_link_type_updated() called in init but don't want to create
            # the default overlay yet, since center_on_viewer() will cause a traceback
            if not center_the_overlay:
                return
            # create new entry with defaults (any defaults not provided here will be carried over
            # from the previous selection based on current traitlet values)

            # if this is the first overlay, there is a chance the traitlet for color was already set
            # to something other than the default, so we want to use that, otherwise for successive
            # new overlays, we want to ignore the traitlet and default back to "active" orange.
            default_color = '#c75109' if len(self._overlays) else self.color
            self._overlays[overlay_selected] = {'color': default_color}

            # similarly, if the user called any import APIs before opening the plugin, we want to
            # respect that, but when creating successive overlays, any selection from file/region
            # should be cleared for the next selection
            if self.preset_selected == 'From File...' and len(self._overlays) > 1:
                self._overlays[overlay_selected]['from_file'] = ''
                self._overlays[overlay_selected]['preset'] = self.preset.choices[0]

            # for the first overlay only, default the position to be centered on the current
            # zoom limits of the first selected viewer
            if len(self._overlays) == 1 and len(self.viewer.selected):
                self.center_on_viewer(self.viewer.selected[0])

        fp = self._overlays[overlay_selected]

        # we'll temporarily disable updating the overlays so that we can set all
        # traitlets simultaneously (and since we're only updating traitlets to a previously-set
        # overlay, we shouldn't have to update anything with the marks themselves)
        self._ignore_traitlet_change = True
        for attr in ('from_file', 'preset_selected',
                     'visible', 'color', 'fill_opacity', 'viewer_selected',
                     'ra', 'dec', 'pa', 'v2_offset', 'v3_offset'):
            key = attr.split('_selected')[0]

            if key in fp:
                # then take the value previously set in the dictionary and apply it to the traitlet
                # (which corresponds to the current selection).
                setattr(self, attr, fp[key])
            else:
                # then adopt the value from the traitlet (copy from the previous selection).
                # This should only ever occur for entries that are not defined in the default
                # dict above.
                fp[key] = getattr(self, attr)
        self._ignore_traitlet_change = False
        self._preset_args_changed(overlay_selected=overlay_selected)

    def _mark_visible(self, viewer_id, overlay=None):
        if not self.is_active:
            return False
        if self.is_pixel_linked:
            return False
        if overlay is None:
            overlay = self.overlay_selected
        if overlay not in self._overlays:
            return False
        fp = self._overlays[overlay]
        if viewer_id not in fp.get('viewer', []):
            return False
        else:
            return fp.get('visible', False)

    @observe('viewer_selected', 'visible', 'color', 'fill_opacity')
    def _display_args_changed(self, msg={}):
        if self._ignore_traitlet_change:
            return
        if not self.overlay_selected:
            return
        if self.overlay_selected not in self._overlays:
            # default dictionary has not been created yet
            return
        self._ensure_first_overlay()
        name = msg.get('name', '').split('_selected')[0]
        if len(name):
            self._overlays[self.overlay_selected][name] = msg.get('new')
        if not self.is_active:
            return

        # update existing mark (or add/remove from viewers)
        for viewer_id, viewer in self.app._viewer_store.items():
            visible = self._mark_visible(viewer_id)
            existing_marks = self._get_marks(viewer, self.overlay_selected)

            if visible and not len(existing_marks):
                # then we need to create new marks at least for this viewer
                self._preset_args_changed()
                return

            for mark in existing_marks:
                mark.visible = visible
                mark.colors = [self.color]
                mark.fill_opacities = [self.fill_opacity]

    def import_region(self, region):
        """
        Import an Astropy regions object (or file).

        Parameters
        ----------
        region : str or regions.Regions object
        """
        self._ensure_first_overlay()
        if isinstance(region, (regions.Region, regions.Regions)):
            self.preset.import_obj(region)
        elif isinstance(region, str):  # TODO: support path objects?
            self.preset.import_file(region)
        else:
            raise TypeError("region must be a regions.Regions object or string (file path)")
        # _preset_args_changed was probably already triggered by from_file traitlet changing, but
        # that may have been before the file was fully parsed and available from preset.selected_obj
        self._preset_args_changed()

    @property
    def overlay_regions(self):
        """
        Access the regions objects corresponding to the current settings
        """

        callable_kwargs = {k: float(getattr(self, k))
                           for k in ('ra', 'dec', 'pa', 'v2_offset', 'v3_offset')}

        if self.preset_selected == 'From File...':
            # we need to cache these locally in order to support multiple files/regions between
            # different overlay entries all selecting From File...
            overlay = self._overlays.get(self.overlay_selected, {})
            if ('regions' not in overlay
                    and isinstance(self.preset.selected_obj, (regions.Region, regions.Regions))):
                regs = self.preset.selected_obj
                if not isinstance(regs, regions.Regions):
                    # then this is a single region, but to be compatible with looping logic,
                    # let's just put as a single entry in a list
                    regs = [regs]
                overlay['regions'] = regs
            regs = overlay.get('regions', [])
        elif self.has_pysiaf:
            regs = preset_regions.instrument_footprint(
                self.preset.selected_item['observatory'],
                self.preset_selected, **callable_kwargs
            )
        else:  # pragma: no cover
            regs = []
        return regs

    @observe('preset_obs_selected')
    def _update_preset_filters(self, event={}):
        if not hasattr(self, 'preset'):
            # during plugin init
            return

        def only_jwst(item):
            return item['label'] == 'From File...' or item.get('observatory') == 'JWST'

        def only_roman(item):
            return item['label'] == 'From File...' or item.get('observatory') == 'Roman'

        if self.preset_obs_selected == 'JWST':
            self.preset.filters = [only_jwst]
        elif self.preset_obs_selected == 'Roman':
            self.preset.filters = [only_roman]
        else:
            self.preset.filters = []

    @observe('preset_selected', 'from_file', 'ra', 'dec', 'pa', 'v2_offset', 'v3_offset')
    def _preset_args_changed(self, msg={}, overlay_selected=None):
        if self._ignore_traitlet_change:
            return
        overlay_selected = overlay_selected if overlay_selected is not None else self.overlay_selected  # noqa
        if not overlay_selected:
            return

        name = msg.get('name', '').split('_selected')[0]

        if overlay_selected not in self._overlays:
            # default dictionary has not been created yet
            return

        if len(name):
            self._overlays[overlay_selected][name] = msg.get('new')
        if name == 'from_file' and 'regions' in self._overlays[overlay_selected]:
            # then the file may have been changed from the API, so we need to clear the cache
            # the cache will then be re-populated on the call to self.overlay_regions below
            del self._overlays[overlay_selected]['regions']

        regs = self.overlay_regions
        for viewer_id, viewer in self.app._viewer_store.items():
            visible = self._mark_visible(viewer_id)
            # TODO: need to re-call this logic when the reference_data is changed... which might
            # warrant some refactoring so we don't have to loop over all viewers if it has only
            # changed in one viewer
            wcs = getattr(viewer.state.reference_data, 'coords', None)
            if wcs is None:
                continue
            existing_overlays = self._get_marks(viewer, overlay_selected)
            update_existing = len(existing_overlays) == len(regs)
            if not update_existing and len(existing_overlays):
                # clear any existing marks (length has changed, perhaps a new preset)
                viewer.figure.marks = [m for m in viewer.figure.marks
                                       if getattr(m, 'overlay', None) != overlay_selected]

            # the following logic is adapted from
            # https://github.com/spacetelescope/jwst_novt/blob/main/jwst_novt/interact/display.py
            new_marks = []
            for i, reg in enumerate(regs):
                if (not isinstance(reg, regions.Region)
                        or not hasattr(reg, 'to_pixel')):   # pragma: no cover
                    # NOTE: this is pre-checked for API/file selection in the file-parser
                    # and built-in presets should be designed to never hit this error
                    # in the future we may support pixel regions as well, but need to decide how
                    # to properly handle those scenarios for both WCS and pixel-linking
                    raise NotImplementedError("regions must all be SkyRegions")

                pixel_region = reg.to_pixel(wcs)

                if isinstance(pixel_region, regions.PolygonPixelRegion):
                    x_coords = pixel_region.vertices.x
                    y_coords = pixel_region.vertices.y

                # bqplot marker does not respect image pixel sizes, so need to render as polygon.
                elif isinstance(pixel_region, regions.RectanglePixelRegion):
                    pixel_region = pixel_region.to_polygon()
                    x_coords = pixel_region.vertices.x
                    y_coords = pixel_region.vertices.y
                elif isinstance(pixel_region, (regions.CirclePixelRegion,
                                               regions.EllipsePixelRegion,
                                               regions.CircleAnnulusPixelRegion)):
                    roi = regions2roi(pixel_region)
                    x_coords, y_coords = roi.to_polygon()
                else:  # pragma: no cover
                    raise NotImplementedError("could not parse coordinates from regions - please report this issue")  # noqa

                if update_existing:
                    mark = existing_overlays[i]
                    with mark.hold_sync():
                        mark.x = x_coords
                        mark.y = y_coords
                else:
                    mark = FootprintOverlay(
                        viewer,
                        overlay_selected,
                        x=x_coords,
                        y=y_coords,
                        colors=[self.color],
                        fill_opacities=[self.fill_opacity],
                        visible=visible)
                    new_marks.append(mark)

            if not update_existing and len(new_marks):
                viewer.figure.marks = viewer.figure.marks + new_marks



from contextlib import contextmanager
from echo import delay_callback, CallbackProperty
import numpy as np

from astropy import units as u
from glue.viewers.profile.state import ProfileViewerState
from glue_jupyter.bqplot.image.state import BqplotImageViewerState
from glue.viewers.matplotlib.state import DeferredDrawCallbackProperty as DDCProperty

from jdaviz.utils import get_reference_image_data
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               flux_conversion_general)

__all__ = ['FreezableState', 'FreezableProfileViewerState', 'FreezableBqplotImageViewerState']


class FreezableState:
    _frozen_state = []

    def __setattr__(self, k, v):
        if k[0] == '_' or k not in self._frozen_state:
            super().__setattr__(k, v)
        elif getattr(self, k) is None:
            # still allow Nones to be updated to initial values
            super().__setattr__(k, v)


class FreezableProfileViewerState(ProfileViewerState, FreezableState):
    show_uncertainty = DDCProperty(False, docstring='Whether to show data uncertainties')

    def _reset_x_limits(self, *event):
        # override glue's _reset_x_limits to account for all layers,
        # not just reference data (_reset_y_limits already does so)
        # This is essentially copied directly from Glue's
        # ProfileViewerState._reset_y_limits but modified for x-limits
        if self.reference_data is None or self.x_att_pixel is None:
            return

        x_min, x_max = np.inf, -np.inf
        for layer in self.layers:
            try:
                profile = layer.profile
            except Exception:  # nosec
                # e.g. incompatible subset
                continue
            if profile is not None:
                x, y = profile
                if len(x) > 0:
                    x_min = min(x_min, np.nanmin(x))
                    x_max = max(x_max, np.nanmax(x))

        if not np.all(np.isfinite([x_min, x_max])):
            return

        with delay_callback(self, 'x_min', 'x_max'):
            self.x_min = x_min
            self.x_max = x_max

    def _convert_units_y_limits(self, old_unit, new_unit):
        # override glue's _convert_units_y_limits to account
        # for equivalencies.  This converts all four corners
        # of the limits to set new limits that contain those
        # same corners

        if old_unit != new_unit:
            if old_unit is None or new_unit is None:
                self._reset_y_limits()
                return
            x_corners = np.array([self.x_min, self.x_min, self.x_max, self.x_max])
            y_corners = np.array([self.y_min, self.y_max, self.y_min, self.y_max])
            spectral_axis = x_corners * u.Unit(self.x_display_unit)

            # NOTE: this uses the scale factor from the first found layer.  We may want to
            # generalize this to iterate over all scale factors if/when we support multiple
            # flux cubes (with potentially different pixel scale factors).
            eqv = None
            for layer in self.layers:
                if psc := getattr(layer.layer, 'meta', {}).get('_pixel_scale_factor', None):  # noqa
                    spectral_axis.info.meta = {'_pixel_scale_factor',
                                               psc}
                    eqv = all_flux_unit_conversion_equivs(pixar_sr=psc,
                                                          cube_wave=spectral_axis)
                    break
            else:
                spectral_axis.info.meta = {}
                eqv = all_flux_unit_conversion_equivs(cube_wave=spectral_axis)
                spectral_axis = None

            y_corners_new = flux_conversion_general(y_corners, old_unit, new_unit, eqv, with_unit=False)  # noqa

            with delay_callback(self, 'y_min', 'y_max'):
                self.y_min = np.nanmin(y_corners_new)
                self.y_max = np.nanmax(y_corners_new)


class FreezableBqplotImageViewerState(BqplotImageViewerState, FreezableState):
    linked_by_wcs = False

    zoom_radius = CallbackProperty(1.0, docstring="Zoom radius")
    zoom_center_x = CallbackProperty(0.0, docstring='x-coordinate of center of zoom box')
    zoom_center_y = CallbackProperty(0.0, docstring='y-coordinate of center of zoom box')

    def __init__(self, *args, **kwargs):
        self.wcs_only_layers = []  # For Imviz rotation use.
        self._during_zoom_sync = False
        self.add_callback('zoom_radius', self._set_zoom_radius_center)
        self.add_callback('zoom_center_x', self._set_zoom_radius_center)
        self.add_callback('zoom_center_y', self._set_zoom_radius_center)
        for attr in ('x_min', 'x_max', 'y_min', 'y_max'):
            self.add_callback(attr, self._set_axes_lim)
        super().__init__(*args, **kwargs)

    def _set_viewer(self, viewer):
        self._viewer = viewer
        self._set_axes_lim()

    @contextmanager
    def during_zoom_sync(self):
        self._during_zoom_sync = True
        try:
            yield
        except Exception:
            self._during_zoom_sync = False
            raise
        self._during_zoom_sync = False

    def _set_zoom_radius_center(self, *args):
        if self._during_zoom_sync or not hasattr(self, '_viewer'):
            return

        # When WCS-linked (displayed on the sky): zoom_center_x/y and zoom_radius are in sky units,
        # x/y_min/max are in pixels of the WCS-only layer
        if self.linked_by_wcs:
            image, i_ref = get_reference_image_data(self._viewer.jdaviz_app, self._viewer.reference)
            ref_wcs = image.coords
            cr = ref_wcs.world_to_pixel_values((self.zoom_center_x, self.zoom_center_x+abs(self.zoom_radius)),  # noqa
                                               (self.zoom_center_y, self.zoom_center_y))
            center_x, center_xr = cr[0]
            center_y, _ = cr[1]
            radius = abs(center_xr - center_x)
        else:
            center_x, center_y = self.zoom_center_x, self.zoom_center_y
            radius = abs(self.zoom_radius)
        # now center_x/y and radius are in pixel units of the reference data, so can be used to
        # update limits

        with self.during_zoom_sync():
            x_min = center_x - radius
            x_max = center_x + radius
            y_min = center_y - radius
            y_max = center_y + radius
            self.x_min, self.x_max, self.y_min, self.y_max = x_min, x_max, y_min, y_max

            self._adjust_limits_aspect()

    def _set_axes_aspect_ratio(self, axes_ratio):
        # when aspect-ratio is changed (changing viewer.shape), ensure zoom/center are synced
        # with zoom-limits
        super()._set_axes_aspect_ratio(axes_ratio)
        self._set_axes_lim()

    def _set_axes_lim(self, *args):
        if self._during_zoom_sync or not hasattr(self, '_viewer'):
            return
        if None in (self.x_min, self.x_max, self.y_min, self.y_max):
            return

        # When WCS-linked (displayed on the sky): zoom_center_x/y and zoom_radius are in sky units,
        # x/y_min/max are in pixels of the WCS-only layer
        if self.linked_by_wcs:
            image, i_ref = get_reference_image_data(self._viewer.jdaviz_app, self._viewer.reference)
            ref_wcs = image.coords
            lims = ref_wcs.pixel_to_world_values((self.x_min, self.x_max), (self.y_min, self.y_max))
            x_min, x_max = lims[0]
            y_min, y_max = lims[1]
        else:
            x_min, y_min = self.x_min, self.y_min
            x_max, y_max = self.x_max, self.y_max
        # now x_min/max, y_min/max are in axes units (degrees if WCS-linked, pixels otherwise)

        with self.during_zoom_sync():
            self.zoom_radius = abs(0.5 * min(x_max - x_min, y_max - y_min))
            self.zoom_center_x = 0.5 * (x_max + x_min)
            self.zoom_center_y = 0.5 * (y_max + y_min)

    def _get_reset_limits(self, return_as_world=False):
        wcs_success = False
        if self.linked_by_wcs and self.reference_data.coords is not None:
            x_min, x_max = np.inf, -np.inf
            y_min, y_max = np.inf, -np.inf

            for layer in self.layers:
                if not layer.visible:
                    continue

                data = next((x for x in self.data_collection if x.label == layer.layer.data.label))
                if data.coords is None:
                    # if no layers have coords, then wcs_success will remain
                    # false and limits will fallback based on pixel limit
                    continue

                pixel_ids = layer.layer.pixel_component_ids
                world_bottom_left = data.coords.pixel_to_world(0, 0)
                world_top_right = data.coords.pixel_to_world(
                    layer.layer.data.shape[pixel_ids[1].axis] - 1,
                    layer.layer.data.shape[pixel_ids[0].axis] - 1
                )

                if return_as_world:
                    x_min = min(x_min, world_bottom_left.ra.value)
                    x_max = max(x_max, world_top_right.ra.value)
                    y_min = min(y_min, world_bottom_left.dec.value)
                    y_max = max(y_max, world_top_right.dec.value)
                else:
                    x_bl, y_bl = self.reference_data.coords.world_to_pixel(world_bottom_left)
                    x_tr, y_tr = self.reference_data.coords.world_to_pixel(world_top_right)

                    x_pix_min = min(x_bl, x_tr)
                    x_pix_max = max(x_bl, x_tr)
                    y_pix_min = min(y_bl, y_tr)
                    y_pix_max = max(y_bl, y_tr)

                    x_min = min(x_min, x_pix_min - 0.5)
                    x_max = max(x_max, x_pix_max + 0.5)
                    y_min = min(y_min, y_pix_min - 0.5)
                    y_max = max(y_max, y_pix_max + 0.5)
                wcs_success = True

        if not wcs_success:
            x_min, x_max = -0.5, -np.inf
            y_min, y_max = -0.5, -np.inf
            for layer in self.layers:
                if not layer.visible or layer.layer.data.ndim == 1:
                    continue
                pixel_ids = layer.layer.pixel_component_ids
                pixel_id_x = [comp for comp in pixel_ids if comp.label.endswith('[x]')][0]
                pixel_id_y = [comp for comp in pixel_ids if comp.label.endswith('[y]')][0]

                x_max = max(x_max, layer.layer.data.shape[pixel_id_x.axis] - 0.5)
                y_max = max(y_max, layer.layer.data.shape[pixel_id_y.axis] - 0.5)

        return x_min, x_max, y_min, y_max

    def reset_limits(self, *event):
        # TODO: use consistent logic for all image viewers by removing this if-statement
        # if/when WCS linking is supported (i.e. in cubeviz)
        if getattr(self, '_viewer', None) is not None and self._viewer.jdaviz_app.config != 'imviz':
            return super().reset_limits(*event)
        if self.reference_data is None:  # Nothing to do
            return

        x_min, x_max, y_min, y_max = self._get_reset_limits()

        # If any bound wasn't set to a real value, don't update
        if np.any(~np.isfinite([x_min, x_max, y_min, y_max])):
            return

        with delay_callback(self, 'x_min', 'x_max', 'y_min', 'y_max'):
            self.x_min, self.x_max, self.y_min, self.y_max = x_min, x_max, y_min, y_max
            # We need to adjust the limits in here to avoid triggering all
            # the update events then changing the limits again.
            self._adjust_limits_aspect()



import numpy as np

from traitlets import Bool, List, Unicode, observe

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        SelectPluginComponent,
                                        DatasetSelect,
                                        AddResults,
                                        skip_if_no_updates_since_last_active,
                                        skip_if_not_tray_instance,
                                        with_spinner)
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.custom_traitlets import IntHandleEmpty, FloatHandleEmpty
from jdaviz.core.marks import PluginLine

from astropy.modeling import models
from astropy.nddata import StdDevUncertainty, VarianceUncertainty, UnknownUncertainty
from specreduce import tracing
from specreduce import background
from specreduce import extract

__all__ = ['SpectralExtraction']

_model_cls = {'Spline': models.Spline1D,
              'Polynomial': models.Polynomial1D,
              'Legendre': models.Legendre1D,
              'Chebyshev': models.Chebyshev1D}


@tray_registry('spectral-extraction', label="Spectral Extraction",
               viewer_requirements=['spectrum', 'spectrum-2d'])
class SpectralExtraction(PluginTemplateMixin):
    """
    The Spectral Extraction plugin exposes specreduce methods for tracing, background subtraction,
    and spectral extraction from 2D spectra.
    See the :ref:`Spectral Extraction Plugin Documentation <specviz2d-spectral-extraction>`
    for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * :attr:`interactive_extract`
    * ``trace_dataset`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      controls the input dataset for generating the trace.
    * ``trace_type`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      controls the type of trace to be generated.
    * ``trace_peak_method`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      only applicable if ``trace_type`` is not ``Flat``.
    * :attr:`trace_pixel` :
      pixel of the trace.  If ``trace_type`` is not ``Flat``, then this
      is the "guess" for the automated trace.
    * :attr:`trace_do_binning` :
      only applicable if ``trace_type`` is not ``Flat``.  Bin the input data when fitting the
      trace.
    * :attr:`trace_bins` :
      only applicable if ``trace_type`` is not ``Flat`` and ``trace_do_binning``.
    * :attr:`trace_window` :
      full width of the trace.
    * :meth:`import_trace`
    * :meth:`export_trace`
    * ``bg_dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      controls the input dataset for generating the background.
    * ``bg_type`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      controls the type of background to be generated.
    * :attr:`bg_trace_pixel` :
      only applicable if ``bg_type`` is set to ``Manual``
    * :attr:`bg_separation` :
      only applicable if ``bg_type`` set set to ``OneSided`` or ``TwoSided``.
      Separation from the referenced trace for the center of each of the background window(s).
    * :attr:`bg_width` :
      full width of each background window(s).
    * ``bg_statistic`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      statistic to use when computing the background.  'Average' will account for partial pixel
      weights, 'Median' will include all partial pixels.
    * ``bg_add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * ``bg_sub_add_results``
    * :meth:`import_bg`
    * :meth:`export_bg`
    * :meth:`export_bg_spectrum`
    * :meth:`export_bg_img`
    * :meth:`export_bg_sub`
    * ``ext_dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      controls the input dataset for generating the extracted spectrum.  "From Plugin" will
      use the background-subtracted image as defined by the background options above.  To skip
      background extraction, select the original input 2D spectrum.
    * ``ext_trace`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`)
    * ``ext_type`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * :attr:`ext_width` :
      full width of the extraction window.
    * ``horne_ext_profile`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      For Horne extract, choice of 'Gaussian' or 'Self (interpolated)' to use
      empirical profile from data.
    * :attr:`self_prof_n_bins` :
      Number of bins to use when computing the self-derived profile for Horne Extract.
    * :attr:`self_prof_interp_degree_x` :
      Interpolation degree (in X) to use when computing the self-derived profile
      for Horne Extract.
    * :attr:`self_prof_interp_degree_y` :
      Interpolation degree (in Y) to use when computing the self-derived profile
      for Horne Extract.
    * ``ext_add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`import_extract`
    * :meth:`export_extract`
    * :meth:`export_extract_spectrum`
    """
    dialog = Bool(False).tag(sync=True)
    template_file = __file__, "spectral_extraction.vue"
    uses_active_status = Bool(True).tag(sync=True)

    active_step = Unicode().tag(sync=True)
    interactive_extract = Bool(True).tag(sync=True)

    # TRACE
    trace_trace_items = List().tag(sync=True)
    trace_trace_selected = Unicode().tag(sync=True)

    trace_offset = IntHandleEmpty(0).tag(sync=True)

    trace_dataset_items = List().tag(sync=True)
    trace_dataset_selected = Unicode().tag(sync=True)

    trace_type_items = List().tag(sync=True)
    trace_type_selected = Unicode().tag(sync=True)

    trace_pixel = FloatHandleEmpty(0).tag(sync=True)
    trace_order = IntHandleEmpty(3).tag(sync=True)

    trace_peak_method_items = List().tag(sync=True)
    trace_peak_method_selected = Unicode().tag(sync=True)

    trace_do_binning = Bool(True).tag(sync=True)
    trace_bins = IntHandleEmpty(20).tag(sync=True)
    trace_window = IntHandleEmpty(0).tag(sync=True)

    trace_results_label = Unicode().tag(sync=True)
    trace_results_label_default = Unicode().tag(sync=True)
    trace_results_label_auto = Bool(True).tag(sync=True)
    trace_results_label_invalid_msg = Unicode('').tag(sync=True)
    trace_results_label_overwrite = Bool().tag(sync=True)
    trace_add_to_viewer_items = List().tag(sync=True)
    trace_add_to_viewer_selected = Unicode().tag(sync=True)
    trace_spinner = Bool(False).tag(sync=True)

    # BACKGROUND
    bg_dataset_items = List().tag(sync=True)
    bg_dataset_selected = Unicode().tag(sync=True)

    bg_type_items = List().tag(sync=True)
    bg_type_selected = Unicode().tag(sync=True)

    bg_trace_items = List().tag(sync=True)
    bg_trace_selected = Unicode().tag(sync=True)

    bg_trace_pixel = FloatHandleEmpty(0).tag(sync=True)

    bg_statistic_items = List().tag(sync=True)
    bg_statistic_selected = Unicode().tag(sync=True)

    bg_separation = FloatHandleEmpty(0).tag(sync=True)
    bg_width = FloatHandleEmpty(0).tag(sync=True)

    bg_results_label = Unicode().tag(sync=True)
    bg_results_label_default = Unicode().tag(sync=True)
    bg_results_label_auto = Bool(True).tag(sync=True)
    bg_results_label_invalid_msg = Unicode('').tag(sync=True)
    bg_results_label_overwrite = Bool().tag(sync=True)
    bg_add_to_viewer_items = List().tag(sync=True)
    bg_add_to_viewer_selected = Unicode().tag(sync=True)
    bg_img_spinner = Bool(False).tag(sync=True)

    bg_spec_results_label = Unicode().tag(sync=True)
    bg_spec_results_label_default = Unicode().tag(sync=True)
    bg_spec_results_label_auto = Bool(True).tag(sync=True)
    bg_spec_results_label_invalid_msg = Unicode('').tag(sync=True)
    bg_spec_results_label_overwrite = Bool().tag(sync=True)
    bg_spec_add_to_viewer_items = List().tag(sync=True)
    bg_spec_add_to_viewer_selected = Unicode().tag(sync=True)
    bg_spec_spinner = Bool(False).tag(sync=True)

    bg_sub_results_label = Unicode().tag(sync=True)
    bg_sub_results_label_default = Unicode().tag(sync=True)
    bg_sub_results_label_auto = Bool(True).tag(sync=True)
    bg_sub_results_label_invalid_msg = Unicode('').tag(sync=True)
    bg_sub_results_label_overwrite = Bool().tag(sync=True)
    bg_sub_add_to_viewer_items = List().tag(sync=True)
    bg_sub_add_to_viewer_selected = Unicode().tag(sync=True)
    bg_sub_spinner = Bool(False).tag(sync=True)

    # EXTRACT
    ext_dataset_items = List().tag(sync=True)
    ext_dataset_selected = Unicode().tag(sync=True)

    ext_trace_items = List().tag(sync=True)
    ext_trace_selected = Unicode().tag(sync=True)

    ext_type_items = List().tag(sync=True)
    ext_type_selected = Unicode().tag(sync=True)

    horne_ext_profile_items = List().tag(sync=True)
    horne_ext_profile_selected = Unicode().tag(sync=True)

    self_prof_n_bins = IntHandleEmpty(10).tag(sync=True)
    self_prof_interp_degree_x = IntHandleEmpty(1).tag(sync=True)
    self_prof_interp_degree_y = IntHandleEmpty(1).tag(sync=True)

    ext_width = FloatHandleEmpty(0).tag(sync=True)

    ext_uncert_warn = Bool(False).tag(sync=True)
    ext_specreduce_err = Unicode().tag(sync=True)

    ext_results_label = Unicode().tag(sync=True)
    ext_results_label_default = Unicode().tag(sync=True)
    ext_results_label_auto = Bool(True).tag(sync=True)
    ext_results_label_invalid_msg = Unicode('').tag(sync=True)
    ext_results_label_overwrite = Bool().tag(sync=True)
    ext_add_to_viewer_items = List().tag(sync=True)
    ext_add_to_viewer_selected = Unicode().tag(sync=True)
    # uses default "spinner"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Extract 1D spectrum from 2D image.'

        self._marks = {}

        # TRACE
        self.trace_trace = DatasetSelect(self,
                                         'trace_trace_items',
                                         'trace_trace_selected',
                                         default_text='New Trace',
                                         filters=['is_trace'])

        self.trace_dataset = DatasetSelect(self,
                                           'trace_dataset_items',
                                           'trace_dataset_selected',
                                           filters=['layer_in_spectrum_2d_viewer', 'not_trace'])

        self.trace_type = SelectPluginComponent(self,
                                                items='trace_type_items',
                                                selected='trace_type_selected',
                                                manual_options=['Flat', 'Polynomial',
                                                                'Legendre', 'Chebyshev',
                                                                'Spline'])

        self.trace_peak_method = SelectPluginComponent(self,
                                                       items='trace_peak_method_items',
                                                       selected='trace_peak_method_selected',
                                                       manual_options=['Gaussian', 'Centroid', 'Max'])  # noqa

        self.trace_add_results = AddResults(self, 'trace_results_label',
                                            'trace_results_label_default',
                                            'trace_results_label_auto',
                                            'trace_results_label_invalid_msg',
                                            'trace_results_label_overwrite',
                                            'trace_add_to_viewer_items',
                                            'trace_add_to_viewer_selected')
        self.trace_add_results.viewer.filters = ['is_spectrum_2d_viewer']
        self.trace_results_label_default = 'trace'

        # BACKGROUND
        self.bg_dataset = DatasetSelect(self,
                                        'bg_dataset_items',
                                        'bg_dataset_selected',
                                        filters=['layer_in_spectrum_2d_viewer', 'not_trace'])

        self.bg_type = SelectPluginComponent(self,
                                             items='bg_type_items',
                                             selected='bg_type_selected',
                                             manual_options=['TwoSided', 'OneSided', 'Manual'])

        self.bg_trace = DatasetSelect(self,
                                      items='bg_trace_items',
                                      selected='bg_trace_selected',
                                      default_text='From Plugin',
                                      filters=['is_trace'])

        self.bg_statistic = SelectPluginComponent(self,
                                                  items='bg_statistic_items',
                                                  selected='bg_statistic_selected',
                                                  manual_options=['Average', 'Median'])

        self.bg_add_results = AddResults(self, 'bg_results_label',
                                         'bg_results_label_default',
                                         'bg_results_label_auto',
                                         'bg_results_label_invalid_msg',
                                         'bg_results_label_overwrite',
                                         'bg_add_to_viewer_items',
                                         'bg_add_to_viewer_selected')
        self.bg_add_results.viewer.filters = ['is_spectrum_2d_viewer']
        self.bg_results_label_default = 'background'

        self.bg_spec_add_results = AddResults(self, 'bg_spec_results_label',
                                              'bg_spec_results_label_default',
                                              'bg_spec_results_label_auto',
                                              'bg_spec_results_label_invalid_msg',
                                              'bg_spec_results_label_overwrite',
                                              'bg_spec_add_to_viewer_items',
                                              'bg_spec_add_to_viewer_selected')
        self.bg_spec_add_results.viewer.filters = ['is_spectrum_viewer']
        self.bg_spec_results_label_default = 'background-spectrum'

        self.bg_sub_add_results = AddResults(self, 'bg_sub_results_label',
                                             'bg_sub_results_label_default',
                                             'bg_sub_results_label_auto',
                                             'bg_sub_results_label_invalid_msg',
                                             'bg_sub_results_label_overwrite',
                                             'bg_sub_add_to_viewer_items',
                                             'bg_sub_add_to_viewer_selected')
        self.bg_sub_add_results.viewer.filters = ['is_spectrum_2d_viewer']
        self.bg_sub_results_label_default = 'background-subtracted'

        # EXTRACT
        self.ext_dataset = DatasetSelect(self,
                                         'ext_dataset_items',
                                         'ext_dataset_selected',
                                         default_text='From Plugin',
                                         filters=['layer_in_spectrum_2d_viewer', 'not_trace'])

        self.ext_trace = DatasetSelect(self,
                                       items='ext_trace_items',
                                       selected='ext_trace_selected',
                                       default_text='From Plugin',
                                       filters=['is_trace'])

        self.ext_type = SelectPluginComponent(self,
                                              items='ext_type_items',
                                              selected='ext_type_selected',
                                              manual_options=['Boxcar', 'Horne'])

        self.horne_ext_profile = SelectPluginComponent(self,
                                                       items='horne_ext_profile_items',
                                                       selected='horne_ext_profile_selected',
                                                       manual_options=['Gaussian', 'Self (interpolated)'])  # noqa

        self.ext_add_results = AddResults(self, 'ext_results_label',
                                          'ext_results_label_default',
                                          'ext_results_label_auto',
                                          'ext_results_label_invalid_msg',
                                          'ext_results_label_overwrite',
                                          'ext_add_to_viewer_items',
                                          'ext_add_to_viewer_selected')
        self.ext_add_results.viewer.filters = ['is_spectrum_viewer']
        # NOTE: defaults to overwriting original spectrum
        self.ext_add_results.label_whitelist_overwrite = ['Spectrum 1D']
        self.ext_results_label_default = 'Spectrum 1D'

    @property
    def _default_spectrum_viewer_reference_name(self):
        return self.app._jdaviz_helper._default_spectrum_viewer_reference_name

    @property
    def _default_spectrum_2d_viewer_reference_name(self):
        return self.app._jdaviz_helper._default_spectrum_2d_viewer_reference_name

    @property
    def user_api(self):
        return PluginUserApi(self, expose=('interactive_extract',
                                           'trace_dataset', 'trace_type',
                                           'trace_order', 'trace_peak_method',
                                           'trace_pixel',
                                           'trace_do_binning', 'trace_bins', 'trace_window',
                                           'import_trace',
                                           'export_trace',
                                           'bg_dataset', 'bg_type',
                                           'bg_trace_pixel', 'bg_separation', 'bg_width',
                                           'bg_statistic',
                                           'bg_add_results', 'bg_sub_add_results',
                                           'import_bg',
                                           'export_bg', 'export_bg_spectrum',
                                           'export_bg_img', 'export_bg_sub',
                                           'ext_dataset', 'ext_trace', 'ext_type',
                                           'ext_width', 'ext_add_results',
                                           'horne_ext_profile',
                                           'self_prof_n_bins',
                                           'self_prof_interp_degree_x',
                                           'self_prof_interp_degree_y',
                                           'import_extract',
                                           'export_extract', 'export_extract_spectrum'))

    @observe('trace_dataset_selected')
    def _trace_dataset_selected(self, msg=None):
        if not hasattr(self, 'trace_dataset'):
            # happens when first initializing plugin outside of tray
            return

        width = self.trace_dataset.get_selected_spectrum(use_display_units=True).shape[0]
        # estimate the pixel number by taking the median of the brightest pixel index
        # in each column, ignoring columns where the sum in that column is not
        # positive (ie. columns of all zeros or nans)
        trace_flux = self.trace_dataset.get_selected_spectrum(use_display_units=True).flux
        trace_flux_ignore_zeros = trace_flux[:, np.nansum(trace_flux, axis=0) != 0]
        if trace_flux_ignore_zeros.shape[1] == 0:
            # default to trace in middle of image
            brightest_pixel = int(trace_flux.shape[0]/2)
        else:
            brightest_pixel = int(np.nanmedian(np.argmax(trace_flux_ignore_zeros, axis=0)))
        # do not allow to be an edge pixel
        if brightest_pixel < 1:
            brightest_pixel = 1
        if brightest_pixel > width - 1:
            brightest_pixel = width - 1
        distance_from_edge = min(brightest_pixel, width-brightest_pixel)
        # default width will be 10% of cross-dispersion "height",
        # but no larger than distance from the edge
        default_bg_width = int(np.ceil(width / 10))
        default_width = min(default_bg_width, distance_from_edge * 2)

        # sign for one-sided and single trace-pixel depending on whether the brightest pixel is
        # above or below the middle of the image
        if default_bg_width * 2 >= distance_from_edge:
            sign = 1 if (brightest_pixel < width / 2) else -1
            default_bg_separation = sign * default_bg_width * 2
        else:
            default_bg_separation = default_bg_width * 2

        if self.trace_pixel == 0:
            self.trace_pixel = brightest_pixel
        if self.trace_window == 0:
            self.trace_window = default_width
        if self.bg_trace_pixel == 0:
            self.bg_trace_pixel = brightest_pixel + default_bg_separation
        if self.bg_separation == 0:
            if default_bg_width * 2 >= distance_from_edge:
                self.bg_type_selected = 'OneSided'
            self.bg_separation = default_bg_separation
        if self.bg_width == 0:
            self.bg_width = default_bg_width
        if self.ext_width == 0:
            self.ext_width = default_width

    def update_marks(self, step=None):
        """
        Manually update the live-preview marks for a given step in spectral extraction.  This API
        mimics opening the plugin and interacting with one of the steps.

        Parameters
        ----------
        step : str
            Step in the extraction process to visualize.  Must be one of: 'trace', 'bg', 'ext'.
        """
        if step is not None:
            if step == 'trace':
                self._interaction_in_trace_step()
            elif step == 'bg':
                self._interaction_in_bg_step()
            elif step == 'ext':
                self._interaction_in_ext_step()
            else:
                raise ValueError("step must be one of: trace, bg, ext")

    @observe('is_active', 'active_step')
    @skip_if_not_tray_instance()
    def _update_plugin_marks(self, msg={}):
        if self.app._jdaviz_helper is None:
            return
        if not len(self._marks):
            # plugin has never been opened, no need to create marks just to hide them,
            # we'll create marks when the plugin is first opened
            return

        if not (self.is_active):
            for step, mark in self.marks.items():
                mark.visible = False
            return

        if self.active_step == '':
            # on load, default to 'extract' (this will then trigger the observe to update the marks)
            self.active_step = 'ext'
            return

        display_marks = {'trace': ['trace', 'extract'],
                         'bg': ['trace',
                                'bg1_center', 'bg1_lower', 'bg1_upper',
                                'bg2_center', 'bg2_lower', 'bg2_upper',
                                'bg_spec', 'extract'],
                         'ext': ['trace',
                                 'ext_upper', 'ext_lower',
                                 'extract']}
        for step, mark in self.marks.items():
            mark.visible = step in display_marks.get(self.active_step, [])

    @property
    def marks(self):
        """
        Access the marks created by this plugin in both the spectrum-viewer and spectrum-2d-viewer.
        """
        if self._marks:
            # TODO: replace with cache property?
            return self._marks

        if not self._tray_instance:
            return {}

        viewer2d = self.app.get_viewer(self._default_spectrum_2d_viewer_reference_name)
        viewer1d = self.app.get_viewer(self._default_spectrum_viewer_reference_name)

        if not viewer2d.state.reference_data:
            # we don't have data yet for scales, defer initializing
            return {}

        # the marks haven't been initialized yet, so initialize with empty
        # marks that will be populated once the first analysis is done.
        self._marks = {'trace': PluginLine(viewer2d, visible=self.is_active),
                       'ext_lower': PluginLine(viewer2d, visible=self.is_active),
                       'ext_upper': PluginLine(viewer2d, visible=self.is_active),
                       'bg1_center': PluginLine(viewer2d, visible=self.is_active,
                                                line_style='dotted'),
                       'bg1_lower': PluginLine(viewer2d, visible=self.is_active),
                       'bg1_upper': PluginLine(viewer2d, visible=self.is_active),
                       'bg2_center': PluginLine(viewer2d, visible=self.is_active,
                                                line_style='dotted'),
                       'bg2_lower': PluginLine(viewer2d, visible=self.is_active),
                       'bg2_upper': PluginLine(viewer2d, visible=self.is_active)}
        # NOTE: += won't trigger the figure to notice new marks
        viewer2d.figure.marks = viewer2d.figure.marks + list(self._marks.values())

        self._marks['extract'] = PluginLine(viewer1d, visible=self.is_active)
        self._marks['bg_spec'] = PluginLine(viewer1d, visible=self.is_active, stroke_width=1)  # noqa

        # NOTE: += won't trigger the figure to notice new marks
        viewer1d.figure.marks = viewer1d.figure.marks + [self._marks['extract'],
                                                         self._marks['bg_spec']]

        return self._marks

    @observe('interactive_extract')
    @skip_if_no_updates_since_last_active()
    @skip_if_not_tray_instance()
    def _update_interactive_extract(self, event={}):
        # also called by any of the _interaction_in_*_step
        if self.interactive_extract:
            try:
                sp1d = self.export_extract_spectrum(add_data=False)
            except Exception as e:
                # NOTE: ignore error, but will be raised when clicking ANY of the export buttons
                # NOTE: FitTrace or manual background are often giving a
                # "background regions overlapped" error from specreduce
                self.ext_specreduce_err = repr(e)
                self.marks['extract'].clear()
            else:
                self.ext_specreduce_err = ''
                self.marks['extract'].update_xy(sp1d.spectral_axis.value,
                                                sp1d.flux.value)
        else:
            self.marks['extract'].clear()

        if self.interactive_extract and self.active_step == 'bg':
            try:
                spec = self.export_bg_spectrum()
            except Exception:
                self.marks['bg_spec'].clear()
            else:
                self.marks['bg_spec'].update_xy(spec.spectral_axis, spec.flux)
        else:
            self.marks['bg_spec'].clear()

    @observe('is_active', 'trace_dataset_selected', 'trace_type_selected',
             'trace_trace_selected', 'trace_offset', 'trace_order',
             'trace_pixel', 'trace_peak_method_selected',
             'trace_do_binning', 'trace_bins', 'trace_window', 'active_step')
    @skip_if_not_tray_instance()
    @skip_if_no_updates_since_last_active()
    def _interaction_in_trace_step(self, event={}):
        if ((event.get('name', '') in ('active_step', 'is_active') and self.active_step != 'trace')
                or not self.is_active):
            return

        try:
            trace = self.export_trace(add_data=False)
        except Exception:
            # NOTE: ignore error, but will be raised when clicking ANY of the export buttons
            self.marks['trace'].clear()
        else:
            self.marks['trace'].update_xy(range(len(trace.trace)),
                                          trace.trace)
            self.marks['trace'].line_style = 'solid'

        self._update_interactive_extract(event)

        self.active_step = 'trace'

    @observe('is_active', 'bg_dataset_selected', 'bg_type_selected',
             'bg_trace_selected', 'bg_trace_pixel',
             'bg_separation', 'bg_width', 'bg_statistic_selected', 'active_step')
    @skip_if_not_tray_instance()
    @skip_if_no_updates_since_last_active()
    def _interaction_in_bg_step(self, event={}):
        if ((event.get('name', '') in ('active_step', 'is_active') and self.active_step != 'bg')
                or not self.is_active):
            return

        try:
            trace = self._get_bg_trace()
        except Exception:
            # NOTE: ignore error, but will be raised when clicking ANY of the export buttons
            for mark in ['trace', 'bg1_center', 'bg1_lower', 'bg1_upper',
                         'bg2_center', 'bg2_lower', 'bg2_upper', 'bg_spec']:
                self.marks[mark].clear()
        else:
            xs = range(len(trace.trace))
            self.marks['trace'].update_xy(xs,
                                          trace.trace)
            self.marks['trace'].line_style = 'dashed'

            if self.bg_type_selected in ['OneSided', 'TwoSided']:
                self.marks['bg1_center'].update_xy(xs,
                                                   trace.trace+self.bg_separation)
                self.marks['bg1_lower'].update_xy(xs,
                                                  trace.trace+self.bg_separation-self.bg_width/2)
                self.marks['bg1_upper'].update_xy(xs,
                                                  trace.trace+self.bg_separation+self.bg_width/2)
            else:
                self.marks['bg1_center'].clear()
                self.marks['bg1_lower'].update_xy(xs,
                                                  trace.trace-self.bg_width/2)
                self.marks['bg1_upper'].update_xy(xs,
                                                  trace.trace+self.bg_width/2)

            if self.bg_type_selected == 'TwoSided':
                self.marks['bg2_center'].update_xy(xs,
                                                   trace.trace-self.bg_separation)
                self.marks['bg2_lower'].update_xy(xs,
                                                  trace.trace-self.bg_separation-self.bg_width/2)
                self.marks['bg2_upper'].update_xy(xs,
                                                  trace.trace-self.bg_separation+self.bg_width/2)
            else:
                for mark in ['bg2_center', 'bg2_lower', 'bg2_upper']:
                    self.marks[mark].clear()

        self._update_interactive_extract(event)

        self.active_step = 'bg'

    @observe('is_active', 'ext_dataset_selected', 'ext_trace_selected',
             'ext_type_selected', 'ext_width', 'active_step',
             'horne_ext_profile_selected', 'self_prof_n_bins',
             'self_prof_interp_degree_x', 'self_prof_interp_degree_y')
    @skip_if_not_tray_instance()
    @skip_if_no_updates_since_last_active()
    def _interaction_in_ext_step(self, event={}):
        if ((event.get('name', '') in ('active_step', 'is_active') and self.active_step not in ('ext', ''))  # noqa
                or not self.is_active):
            return

        try:
            trace = self._get_ext_trace()
        except Exception:
            # NOTE: ignore error, but will be raised when clicking ANY of the export buttons
            for mark in ['trace', 'ext_lower', 'ext_upper']:
                self.marks[mark].clear()
        else:
            xs = range(len(trace.trace))
            self.marks['trace'].update_xy(xs,
                                          trace.trace)
            self.marks['trace'].line_style = 'dashed'
            if self.ext_type_selected == 'Boxcar':
                self.marks['ext_lower'].update_xy(xs,
                                                  trace.trace-self.ext_width/2)
                self.marks['ext_upper'].update_xy(xs,
                                                  trace.trace+self.ext_width/2)
            else:
                for mark in ['ext_lower', 'ext_upper']:
                    self.marks[mark].clear()

        self._update_interactive_extract(event)

        self.active_step = 'ext'

        # TODO: remove this, the traitlet, and the row in spectral_extraction.vue
        # when specutils handles the warning/exception
        if self.ext_type_selected == 'Horne':
            inp_sp2d = self._get_ext_input_spectrum()
            self.ext_uncert_warn = isinstance(inp_sp2d.uncertainty, UnknownUncertainty)
        else:
            self.ext_uncert_warn = False

    def _set_create_kwargs(self, **kwargs):
        invalid_kwargs = [k for k in kwargs.keys() if not hasattr(self, k)]
        if len(invalid_kwargs):
            raise ValueError(f"{invalid_kwargs} are not valid attributes to pass as kwargs")

        for k, v in kwargs.items():
            setattr(self, k, v)

    def import_trace(self, trace):
        """
        Import the input parameters from an existing specreduce Trace object into the plugin.

        Parameters
        ----------
        trace : specreduce.tracing.Trace
            Trace object to import
        """
        if not isinstance(trace, tracing.Trace):  # pragma: no cover
            raise TypeError("trace must be a specreduce.tracing.Trace object")

        if isinstance(trace, tracing.FlatTrace):
            self.trace_type_selected = 'Flat'
            self.trace_pixel = trace.trace_pos
        elif isinstance(trace, tracing.FitTrace):
            self.trace_type_selected = trace.trace_model.__class__.__name__.strip('1D')
            self.trace_pixel = trace.guess
            self.trace_window = trace.window
            self.trace_bins = trace.bins
            self.trace_do_binning = True
            if hasattr(trace.trace_model, 'degree'):
                self.trace_order = trace.trace_model.degree
        elif isinstance(trace, tracing.ArrayTrace):  # pragma: no cover
            raise NotImplementedError(f"cannot import ArrayTrace into plugin.  Use viz.load_trace instead")  # noqa
        else:  # pragma: no cover
            raise NotImplementedError(f"trace of type {trace.__class__.__name__} not supported")

    @with_spinner('trace_spinner')
    def export_trace(self, add_data=False, **kwargs):
        """
        Create a specreduce Trace object from the input parameters
        defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting trace to the application, according to the options
            defined in the plugin.
        """
        self._set_create_kwargs(**kwargs)
        if len(kwargs) and self.active_step != 'trace':
            self.update_marks(step='trace')

        if self.trace_trace_selected != 'New Trace':
            # then we're offsetting an existing trace
            # for FlatTrace, we can keep and expose a new FlatTrace (which has the advantage of
            # being able to load back into the plugin)
            orig_trace = self.trace_trace.selected_obj
            if isinstance(orig_trace, tracing.FlatTrace):
                trace = tracing.FlatTrace(self.trace_dataset.selected_obj,
                                          orig_trace.trace_pos+self.trace_offset)
            else:
                trace = tracing.ArrayTrace(self.trace_dataset.selected_obj,
                                           self.trace_trace.selected_obj.trace+self.trace_offset)

        elif self.trace_type_selected == 'Flat':
            trace = tracing.FlatTrace(self.trace_dataset.selected_obj,
                                      self.trace_pixel)

        elif self.trace_type_selected in _model_cls:
            trace_model = _model_cls[self.trace_type_selected](degree=self.trace_order)
            trace = tracing.FitTrace(self.trace_dataset.selected_obj,
                                     guess=self.trace_pixel,
                                     bins=int(self.trace_bins) if self.trace_do_binning else None,
                                     window=self.trace_window,
                                     peak_method=self.trace_peak_method_selected.lower(),
                                     trace_model=trace_model)

        else:
            raise NotImplementedError(f"trace_type={self.trace_type_selected} not implemented")

        if add_data:
            self.trace_add_results.add_results_from_plugin(trace, replace=False)

        return trace

    def vue_create_trace(self, *args):
        self.export_trace(add_data=True)

    def _get_bg_trace(self):
        if self.bg_type_selected == 'Manual':
            trace = tracing.FlatTrace(self.trace_dataset.get_selected_spectrum(
                                      use_display_units=True),
                                      self.bg_trace_pixel)
        elif self.bg_trace_selected == 'From Plugin':
            trace = self.export_trace(add_data=False)
        else:
            trace = self.bg_trace.get_selected_spectrum(use_disaply_units=True)

        return trace

    def import_bg(self, bg):
        """
        Import the input parameters from an existing specreduce Background object into the plugin.

        Parameters
        ----------
        bg : specreduce.background.Background
            Background object to import
        """
        if not isinstance(bg, background.Background):  # pragma: no cover
            raise TypeError("bg must be a specreduce.background.Background object")

        # TODO: should we detect/set the referenced dataset?
        trace = self._get_bg_trace()
        if len(bg.traces) == 2:
            # try to detect constant separation
            seps1 = bg.traces[0].trace - trace.trace
            seps2 = trace.trace - bg.traces[1].trace
            if np.all(seps1 == seps1[0]) and np.all(seps2 == seps1[0]):
                self.bg_type_selected = 'TwoSided'
                self.bg_separation = abs(int(seps1[0]))
            else:  # pragma: no cover
                raise NotImplementedError("backgrounds with custom traces not supported (could not detect common separation)")  # noqa
        elif len(bg.traces) == 1:
            # either one_sided or trace, let's see if its constant offset from the trace
            seps = bg.traces[0].trace - trace.trace
            if np.all(seps == seps[0]):
                self.bg_type_selected = 'OneSided'
                self.bg_separation = int(seps[0])
            else:  # pragma: no cover
                raise NotImplementedError("backgrounds with custom traces not supported (could not detect common separation)")  # noqa
        else:  # pragma: no cover
            raise NotImplementedError("backgrounds with more than 2 traces not supported")

        self.bg_width = bg.width

    @with_spinner('bg_spinner')
    def export_bg(self, **kwargs):
        """
        Create a specreduce Background object from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting image to the application, according to the options
            defined in the plugin.
        """
        self._set_create_kwargs(**kwargs)
        if len(kwargs) and self.active_step != 'bg':
            self.update_marks(step='bg')

        trace = self._get_bg_trace()

        if self.bg_type_selected == 'Manual':
            bg = background.Background(self.bg_dataset.get_selected_spectrum(
                                       use_display_units=True),
                                       [trace], width=self.bg_width,
                                       statistic=self.bg_statistic.selected.lower())
        elif self.bg_type_selected == 'OneSided':
            bg = background.Background.one_sided(self.bg_dataset.get_selected_spectrum(
                                                 use_display_units=True),
                                                 trace,
                                                 self.bg_separation,
                                                 width=self.bg_width,
                                                 statistic=self.bg_statistic.selected.lower())
        elif self.bg_type_selected == 'TwoSided':
            bg = background.Background.two_sided(self.bg_dataset.get_selected_spectrum(
                                                 use_display_units=True),
                                                 trace,
                                                 self.bg_separation,
                                                 width=self.bg_width,
                                                 statistic=self.bg_statistic.selected.lower())
        else:  # pragma: no cover
            raise NotImplementedError(f"bg_type={self.bg_type_selected} not implemented")

        return bg

    @with_spinner('bg_img_spinner')
    def export_bg_img(self, add_data=False, **kwargs):
        """
        Create a background 2D spectrum from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting image to the application, according to the options
            defined in the plugin.
        """
        bg_spec = self.export_bg(**kwargs).bkg_image()

        if add_data:
            self.bg_add_results.add_results_from_plugin(bg_spec, replace=True)

        return bg_spec

    def vue_create_bg_img(self, *args):
        try:
            self.export_bg_img(add_data=True)
        except Exception as e:
            self.app.hub.broadcast(
                SnackbarMessage(f"Specreduce background failed with the following error: {repr(e)}",
                                color='error', sender=self)
            )

    @with_spinner('bg_spec_spinner')
    def export_bg_spectrum(self, add_data=False, **kwargs):
        """
        Create a background 1D spectrum from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting spectrum to the application, according to the options
            defined in the plugin.
        """
        spec = self.export_bg(**kwargs).bkg_spectrum()

        if add_data:
            self.bg_spec_add_results.add_results_from_plugin(spec, replace=False)

        return spec

    def vue_create_bg_spec(self, *args):
        self.export_bg_spectrum(add_data=True)

    @with_spinner('bg_sub_spinner')
    def export_bg_sub(self, add_data=False, **kwargs):
        """
        Create a background-subtracted 2D spectrum from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting image to the application, according to the options
            defined in the plugin.
        """
        bg_sub_spec = self.export_bg(**kwargs).sub_image()

        if add_data:
            self.bg_sub_add_results.add_results_from_plugin(bg_sub_spec, replace=True)

        return bg_sub_spec

    def vue_create_bg_sub(self, *args):
        self.export_bg_sub(add_data=True)

    def _get_ext_trace(self):
        if self.ext_trace_selected == 'From Plugin':
            return self.export_trace(add_data=False)
        else:
            return self.ext_trace.get_selected_spectrum(use_display_units=True)

    def _get_ext_input_spectrum(self):
        if self.ext_dataset_selected == 'From Plugin':
            return self.export_bg_sub(add_data=False)
        else:
            return self.ext_dataset.get_selected_spectrum(use_display_units=True)

    def import_extract(self, ext):
        """
        Import the input parameters from an existing specreduce extract object into the plugin.

        Parameters
        ----------
        ext : specreduce.extract.BoxcarExtract
            Extract object to import
        """
        if isinstance(ext, extract.BoxcarExtract):
            self.ext_type_selected = 'Boxcar'
            self.ext_width = ext.width
        elif isinstance(ext, extract.HorneExtract):
            self.ext_type_selected = 'Horne'
        else:  # pragma: no cover
            raise TypeError("ext must be a specreduce.extract.BoxcarExtract or specreduce.extract.HorneExtract object")  # noqa

    def export_extract(self, **kwargs):
        """
        Create a specreduce extraction object from the input parameters defined in the plugin.
        """
        self._set_create_kwargs(**kwargs)
        if len(kwargs) and self.active_step != 'ext':
            self.update_marks(step='ext')

        trace = self._get_ext_trace()
        inp_sp2d = self._get_ext_input_spectrum()

        if self.ext_type_selected == 'Boxcar':
            ext = extract.BoxcarExtract(inp_sp2d, trace, width=self.ext_width)
        elif self.ext_type_selected == 'Horne':
            spatial_profile = None
            if inp_sp2d.uncertainty is None:
                inp_sp2d.uncertainty = VarianceUncertainty(np.ones_like(inp_sp2d.data))

            if not hasattr(inp_sp2d.uncertainty, 'uncertainty_type'):
                inp_sp2d.uncertainty = StdDevUncertainty(inp_sp2d.uncert)

            if self.horne_ext_profile_selected == 'Self (interpolated)':

                # check inputs
                if self.self_prof_n_bins <= 0:
                    raise ValueError('`self_prof_n_bins` must be greater than 0.')
                if self.self_prof_interp_degree_x <= 0:
                    raise ValueError('`self_prof_interp_degree_x` must be greater than 0.')
                if self.self_prof_interp_degree_y <= 0:
                    raise ValueError('`self_prof_interp_degree_y` must be greater than 0.')

                # setup dict of interpolation options
                n_bins_interpolated_profile = self.self_prof_n_bins
                interp_degree = (self.self_prof_interp_degree_x, self.self_prof_interp_degree_y)
                spatial_profile = {'name': 'interpolated_profile',
                                   'n_bins_interpolated_profile': n_bins_interpolated_profile,
                                   'interp_degree': interp_degree}

            elif self.horne_ext_profile_selected == 'Gaussian':
                spatial_profile = 'gaussian'

            else:
                raise ValueError("Horne extraction profile must either be 'Gaussian' or 'Self (interpolated)'")  # noqa

            ext = extract.HorneExtract(inp_sp2d, trace, spatial_profile=spatial_profile)
        else:
            raise NotImplementedError(f"extraction type '{self.ext_type_selected}' not supported")  # noqa

        return ext

    @with_spinner('spinner')
    def export_extract_spectrum(self, add_data=False, **kwargs):
        """
        Create an extracted 1D spectrum from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting spectrum to the application, according to the options
            defined in the plugin.
        """
        extract = self.export_extract(**kwargs)
        spectrum = extract.spectrum

        if add_data:
            self.ext_add_results.add_results_from_plugin(spectrum, replace=False)

        return spectrum

    def vue_extract_spectrum(self, *args):
        self.export_extract_spectrum(add_data=True)



import pytest
from regions import CirclePixelRegion, PixCoord
from jdaviz.core.marks import Lines
from jdaviz.configs.imviz.plugins.parsers import HAS_ROMAN_DATAMODELS


@pytest.mark.skipif(not HAS_ROMAN_DATAMODELS, reason="roman_datamodels is not installed")
def test_previews_roman(rampviz_helper, roman_level_1_ramp):
    _ramp_extraction_previews(rampviz_helper, roman_level_1_ramp)


def test_previews_jwst(rampviz_helper, jwst_level_1b_ramp):
    _ramp_extraction_previews(rampviz_helper, jwst_level_1b_ramp)


def _ramp_extraction_previews(_rampviz_helper, _ramp_file):
    _rampviz_helper.load_data(_ramp_file)

    # add subset:
    region = CirclePixelRegion(center=PixCoord(12.5, 15.5), radius=2)
    _rampviz_helper.plugins['Subset Tools'].import_region(region)
    ramp_extr = _rampviz_helper.plugins['Ramp Extraction']._obj

    subsets = _rampviz_helper.app.get_subsets()
    ramp_cube = _rampviz_helper.app.data_collection[0]
    n_groups = ramp_cube.shape[-1]

    assert len(subsets) == 1
    assert 'Subset 1' in subsets

    integration_viewer = _rampviz_helper.app.get_viewer('integration-viewer')

    # contains a layer for the default ramp extraction and the subset:
    assert len(integration_viewer.layers) == 2

    # profile viewer x-axis is the group dimension
    assert str(integration_viewer.state.x_att) == 'Pixel Axis 2 [x]'

    # no subset previews should be visible yet:
    assert len([
        mark for mark in integration_viewer.native_marks
        # should be a subclass of Lines, should be visible,
        if mark.visible and isinstance(mark, Lines) and
        # and the default profile is a 1D series with length n_groups:
        len(mark.x) == n_groups
    ]) == 1

    # check that when the plugin is active, there's one ramp profile generated by the
    # plugin per pixel in the subset (if show_subset_preview),
    # plus one live preview (if show_live_preview):
    for show_live_preview in [True, False]:
        for show_subset_preview in [True, False]:
            with ramp_extr.as_active():
                ramp_extr.show_live_preview = show_live_preview
                ramp_extr.show_subset_preview = show_subset_preview
                ramp_extr.aperture_selected = 'Subset 1'

                subset_state = subsets[ramp_extr.aperture_selected][0]['subset_state']
                n_pixels_in_subset = subset_state.to_mask(ramp_cube)[..., 0].sum()

                assert len([
                    mark for mark in integration_viewer.custom_marks
                    if mark.visible and isinstance(mark, Lines) and
                    len(mark.x) == n_groups
                ]) == int(show_subset_preview) * n_pixels_in_subset + int(show_live_preview)



from asteval import Interpreter

import multiprocessing as mp
from multiprocessing import Pool

import numpy as np

import astropy.units as u

from specutils import Spectrum1D
from specutils.fitting import fit_lines

__all__ = ['fit_model_to_spectrum']


def fit_model_to_spectrum(spectrum, component_list, expression,
                          run_fitter=False, window=None, n_cpu=None):
    """Fits a `~astropy.modeling.CompoundModel` to a
    `~specutils.Spectrum1D` instance.

    If the input spectrum represents a spectral cube, then fits
    the model to every spaxel in the cube, using
    a multiprocessor pool running in parallel (if ``n_cpu`` is
    larger than 1).

    Parameters
    ----------
    spectrum : `~specutils.Spectrum1D`
        The spectrum to be fitted.

    component_list : list
        Spectral model subcomponents stored in a list.
        Their ``'name'`` attribute must be unique. Each subcomponent
        should be an initialized object from `~astropy.modeling.Model`.

    expression : str
        The arithmetic expression that combines together
        the model subcomponents. The subcomponents are
        referred via their ``'name'`` attribute.

    run_fitter : bool
        **This is currently being ignored for 3D fits.**
        When `False` (the default), the function composes the compound
        model and returns it without fitting.

    window : `None` or `~specutils.SpectralRegion`
        See :func:`specutils.fitting.fit_lines`.

    n_cpu : `None` or int
        **This is only used for spectral cube fitting.**
        Number of cores to use for multiprocessing.
        Using all the cores at once is not recommended.
        If `None`, it will use max cores minus one.
        Set this to 1 for debugging.

    Returns
    -------
    output_model : `~astropy.modeling.CompoundModel` or list
        The model resulting from the fit. In the case of a 1D input
        spectrum, a single model instance is returned. In case of a
        3D spectral cube input, instead o model instances for every
        spaxel, a list with 2D arrays, each one storing fitted parameter
        values for all spaxels, is returned.

    output_spectrum : `~specutils.Spectrum1D`
        The realization of the fitted model as a spectrum. The spectrum
        will be 1D or 3D depending on the shape of input spectrum.
    """
    # Initial guess for the fit.
    initial_model = _build_model(component_list, expression)

    if len(spectrum.shape) > 1:
        return _fit_3D(initial_model, spectrum, window=window, n_cpu=n_cpu)
    else:
        return _fit_1D(initial_model, spectrum, run_fitter, window=window)


def _fit_1D(initial_model, spectrum, run_fitter, filter_non_finite=True, window=None):
    """
    Fits an astropy CompoundModel to a Spectrum1D instance.

    Parameters
    ----------
    initial_model : :class: `astropy.modeling.CompoundModel`
        Initial guess for the model to be fitted.
    spectrum : :class:`specutils.Spectrum1D`
        The spectrum to be fitted.
    run_fitter : bool
        When False (the default), the function composes the compound
        model and returns it without fitting.
    window : `None` or :class:`specutils.spectra.SpectralRegion`
        See :func:`specutils.fitting.fitmodels.fit_lines`.

    Returns
    -------
    output_model : :class: `astropy.modeling.CompoundModel`
        The model resulting from the fit.
    output_spectrum : :class:`specutils.Spectrum1D`
        The realization of the fitted model as a spectrum.

    """
    if run_fitter:
        if spectrum.uncertainty and not np.all(spectrum.uncertainty.array == 0):
            weights = 'unc'
        else:
            weights = None
        output_model = fit_lines(spectrum, initial_model, weights=weights,
                                 filter_non_finite=filter_non_finite, window=window)
        output_values = output_model(spectrum.spectral_axis)
    else:
        # Return without fitting.
        output_model = initial_model
        output_values = initial_model(spectrum.spectral_axis)

    # Build return spectrum
    output_spectrum = Spectrum1D(spectral_axis=spectrum.spectral_axis,
                                 flux=output_values,
                                 mask=spectrum.mask)

    return output_model, output_spectrum


def _fit_3D(initial_model, spectrum, window=None, n_cpu=None):
    """
    Fits an astropy CompoundModel to every spaxel in a cube
    using a multiprocessor pool running in parallel. Computes
    realizations of the models over each spaxel.

    Parameters
    ----------
    initial_model : :class: `astropy.modeling.CompoundModel`
        Initial guess for the model to be fitted.
    spectrum : :class:`specutils.Spectrum1D`
        The spectrum that stores the cube in its 'flux' attribute.
    window : `None` or :class:`specutils.spectra.SpectralRegion`
        See :func:`specutils.fitting.fitmodels.fit_lines`.
    n_cpu : `None` or int
        Number of cores to use for multiprocessing.
        Using all the cores at once is not recommended.
        If `None`, it will use max cores minus one.
        Set this to 1 for debugging.

    Returns
    -------
    output_model : :list: a list that stores 2D arrays. Each array contains one
        parameter from `astropy.modeling.CompoundModel` instances
        fitted to every spaxel in the input cube.
    output_spectrum : :class:`specutils.Spectrum1D`
        The spectrum that stores the fitted model values in its 'flux'
        attribute.
    """
    if n_cpu is None:
        n_cpu = mp.cpu_count() - 1

    # Generate list of all spaxels to be fitted
    spaxels = _generate_spaxel_list(spectrum)

    fitted_models = []

    # Build cube with empty arrays, one per input spaxel. These
    # will store the flux values corresponding to the fitted
    # model realization over each spaxel.
    output_flux_cube = np.zeros(shape=spectrum.flux.shape)

    # Callback to collect results from workers into the cubes
    def collect_result(results):
        for i in range(len(results['x'])):
            x = results['x'][i]
            y = results['y'][i]
            model = results['fitted_model'][i]
            fitted_values = results['fitted_values'][i]

            # Store fitted model parameters
            fitted_models.append({"x": x, "y": y, "model": model})

            # Store fitted values
            output_flux_cube[x, y, :] = fitted_values

    # Run multiprocessor pool to fit each spaxel and
    # compute model values on that same spaxel.
    if n_cpu > 1:
        results = []
        pool = Pool(n_cpu)

        # The communication overhead of spawning a process for each *individual*
        # parameter set is prohibitively high (it's actually faster to run things
        # sequentially). Instead, chunk the spaxel list based on the number of
        # available processors, and have each processor do the model fitting
        # on the entire subset of spaxel tuples, then return the set of results.
        for spx in np.array_split(spaxels, n_cpu):
            # Worker for the multiprocess pool.
            worker = SpaxelWorker(spectrum.flux,
                                  spectrum.spectral_axis,
                                  initial_model,
                                  param_set=spx,
                                  window=window,
                                  mask=spectrum.mask)
            r = pool.apply_async(worker, callback=collect_result)
            results.append(r)
        for r in results:
            r.wait()

        pool.close()

    # This route is only for dev debugging because it is very slow
    # but exceptions will not get swallowed up by multiprocessing.
    else:  # pragma: no cover
        worker = SpaxelWorker(spectrum.flux,
                              spectrum.spectral_axis,
                              initial_model,
                              param_set=spaxels,
                              window=window,
                              mask=spectrum.mask)
        collect_result(worker())

    # Build output 3D spectrum
    funit = spectrum.flux.unit
    output_spectrum = Spectrum1D(wcs=spectrum.wcs,
                                 flux=output_flux_cube * funit,
                                 mask=spectrum.mask)

    return fitted_models, output_spectrum


class SpaxelWorker:
    """
    A class with callable instances that perform fitting over a
    spaxel. It provides the callable for the `Pool.apply_async`
    function, and also holds everything necessary to perform the
    fit over one spaxel.

    Additionally, the callable computes the realization of the
    model just fitted, over that same spaxel. We cannot do these
    two steps (fit and compute) separately, since we cannot
    modify parameter values in an already built CompoundModel
    instance. We need to use the current model instance while
    it still exists.
    """
    def __init__(self, flux_cube, wave_array, initial_model, param_set, window=None, mask=None):
        self.cube = flux_cube
        self.wave = wave_array
        self.model = initial_model
        self.param_set = param_set
        self.window = window
        self.mask = mask

    def __call__(self):
        results = {'x': [], 'y': [], 'fitted_model': [], 'fitted_values': []}

        for parameters in self.param_set:
            x = parameters[0]
            y = parameters[1]

            # Calling the Spectrum1D constructor for every spaxel
            # turned out to be less expensive than expected. Experiments
            # show that the cost amounts to a couple percent additional
            # running time in comparison with a version that uses a 3D
            # spectrum as input. Besides, letting an externally-created
            # spectrum reference into the callable somehow prevents it
            # to execute. This behavior was seen also with other functions
            # passed to the callable.
            flux = self.cube[x, y, :]
            if self.mask is not None:
                mask = self.mask[x, y, :]
            else:
                # If no mask is provided:
                mask = np.zeros_like(flux.value).astype(bool)

            sp = Spectrum1D(spectral_axis=self.wave, flux=flux, mask=mask)

            if sp.uncertainty and not np.all(sp.uncertainty.array == 0):
                weights = 'unc'
            else:
                weights = None
            fitted_model = fit_lines(sp, self.model, window=self.window, weights=weights)

            fitted_values = fitted_model(self.wave)

            results['x'].append(x)
            results['y'].append(y)
            results['fitted_model'].append(fitted_model)
            results['fitted_values'].append(fitted_values)

        return results


def _build_model(component_list, expression):
    """
    Builds an astropy CompoundModel from a list of components
    and an expression that links them to each other.

    Parameters
    ----------
    component_list : list
        Spectral model subcomponents stored in a list.
        Their ``'name'`` attribute must be unique. Each subcomponent
        should be an initialized object from `astropy.modeling.models'
    expression : str
        The arithmetic expression that combines together
        the model subcomponents. The subcomponents are
        referred via their ``'name'`` attribute.

    Returns
    -------
    model : :class:`astropy.modeling.CompoundModel`
        The model resulting from the fit.
    """
    # The expression parser needs the subcomponents stored in a
    # dict, with keys taken from their names. This mechanism can
    # be augmented with status feedback to the UI. This requires
    # the code in here to be refactored to provide the necessary
    # hooks for the GUI (see
    # specviz/specviz/plugins/model_editor/models.py around lines
    # 200-230).

    model_dict = {}

    for component in component_list:
        model_dict[component.name] = component

    aeval = Interpreter(usersyms=model_dict)
    model = aeval(expression)

    return model


def _handle_parameter_units(model, fitted_parameters_cube, param_units):
    """
    Extracts parameter units from a CompoundModel and parameter values
    from a list of 2D numpy arrays, and returns a dict of 2D Quantity
    arrays. The dict values are keyed by the parameter names.

    Parameters
    ----------
    model : :class:`astropy.modeling.CompoundModel`
        An instance of a model.
    fitted_parameters_cube : ndarray
        A 3D array in which the 1st dimension addresses model
        parameters, and the 2nd and 3rd dimensions address
        spaxels.
    param_units : list
        A list with each parameter's units (this is not available
        in the model instance itself).

    Returns
    -------
    fitted_parameters_dict : dict
        2D Quantity arrays keyed by parameter name
    """

    fitted_parameters_dict = {}

    for index in range(len(model.parameters)):
        key = model.param_names[index]
        _ary = fitted_parameters_cube[index, :, :]
        fitted_parameters_dict[key] = u.Quantity(_ary, param_units[index])

    return fitted_parameters_dict


def _generate_spaxel_list(spectrum):
    """
    Generates a list with tuples, each one addressing the (x,y)
    coordinates of a spaxel in a 3-D spectrum cube. If a mask is available,
    skip masked indices.

    Parameters
    ----------
    spectrum : :class:`specutils.Spectrum1D`
        The spectrum that stores the cube in its ``'flux'`` attribute.

    Returns
    -------
    spaxels : list
        List with spaxels
    """
    n_x, n_y, _ = spectrum.flux.shape

    if spectrum.mask is None:
        spx = [[(x, y) for x in range(n_x)] for y in range(n_y)]
    else:
        # return only non-masked spaxels
        spx = [[(x, y) for x in range(n_x) if np.any(~spectrum.mask[x, y])]
               for y in range(n_y) if np.any(~spectrum.mask[:, y])]

    spaxels = [item for sublist in spx for item in sublist]

    return spaxels



from jdaviz.core.events import SliceSelectSliceMessage, NewViewerMessage
from jdaviz.core.helpers import CubeConfigHelper
from jdaviz.configs.rampviz.plugins.viewers import RampvizImageView

__all__ = ['Rampviz']

_temporal_axis_names = ['group', 'groups']


class Rampviz(CubeConfigHelper):
    """Rampviz Helper class"""
    _default_configuration = 'rampviz'
    _default_group_viewer_reference_name = "group-viewer"
    _default_diff_viewer_reference_name = "diff-viewer"
    _default_integration_viewer_reference_name = "integration-viewer"
    _cube_viewer_default_label = _default_group_viewer_reference_name

    _loaded_flux_cube = None
    _cube_viewer_cls = RampvizImageView

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cube_cache = {}

    def load_data(self, data, data_label=None, **kwargs):
        """
        Load and parse a data cube with Cubeviz.
        (Note that only one cube may be loaded per Cubeviz instance.)

        Parameters
        ----------
        data : str, `~roman_datamodels.datamodels.DataModel`, `~astropy.nddata.NDDataArray` or ndarray
            A string file path, Roman DataModel object pointing to the
            data cube, an NDDataArray, or a Numpy array.
            If plain array is given, axes order must be ``(x, y, z)``.
        data_label : str or `None`
            Data label to go with the given data. If not given,
            one will be automatically generated.
        **kwargs : dict
            Extra keywords accepted by Jdaviz application-level parser.
        """  # noqa
        if data_label:
            kwargs['data_label'] = data_label

        super().load_data(data, parser_reference="ramp-data-parser", **kwargs)
        self._set_x_axis()

    def _set_x_axis(self, msg={}):
        group_viewer = self.app.get_viewer(self._default_group_viewer_reference_name)
        ref_data = group_viewer.state.reference_data
        if ref_data:
            group_viewer.state.x_att = ref_data.id["Pixel Axis 0 [z]"]
            group_viewer.state.y_att = ref_data.id["Pixel Axis 1 [y]"]

    def select_group(self, group_index):
        """
        Select the slice closest to the provided wavelength.

        Parameters
        ----------
        group_index : float
            Group index to select in units of the x-axis of the integration.
            The nearest group will be selected if "snap to slice" is enabled
            in the slice plugin.
        """
        if not isinstance(group_index, (int, float)):
            raise TypeError("group_index must be convertible to an integer")
        if group_index < 0:
            raise ValueError("group_index must be positive")

        msg = SliceSelectSliceMessage(value=int(group_index), sender=self)
        self.app.hub.broadcast(msg)

    def get_data(self, data_label=None, spatial_subset=None,
                 temporal_subset=None, cls=None, use_display_units=False):
        """
        Returns data with name equal to ``data_label`` of type ``cls`` with subsets applied from
        ``temporal_subset``, if applicable.

        Parameters
        ----------
        data_label : str, optional
            Provide a label to retrieve a specific data set from data_collection.
        spatial_subset : str, optional
            Spatial subset applied to data.  Only applicable if ``data_label`` points to a cube or
            image.  To extract a spectrum from a cube, use the spectral extraction plugin instead.
        temporal_subset : str, optional
        cls : `~specutils.Spectrum1D`, `~astropy.nddata.CCDData`, optional
            The type that data will be returned as.

        Returns
        -------
        data : cls
            Data is returned as type cls with subsets applied.

        """
        return self._get_data(data_label=data_label, spatial_subset=spatial_subset,
                              temporal_subset=temporal_subset,
                              cls=cls, use_display_units=use_display_units)

    def create_image_viewer(self, viewer_name=None, data=None):
        """
        Create a new image viewer.

        Parameters
        ----------
        viewer_name : str or `None`
            Viewer name/ID to use. If `None`, it is auto-generated.

        Returns
        -------
        viewer : `~jdaviz.configs.imviz.plugins.viewers.ImvizImageView`
            Image viewer instance.

        """
        from jdaviz.configs.rampviz.plugins.viewers import RampvizImageView

        # Cannot assign data to real Data because it loads but it will
        # not update checkbox in Data menu.

        return self.app._on_new_viewer(
            NewViewerMessage(RampvizImageView, data=None, sender=self.app),
            vid=viewer_name, name=viewer_name)



import pytest

# NOTE: Since this is optional dependency, codecov coverage does not include this test module.
roman_datamodels = pytest.importorskip("roman_datamodels")

from gwcs import WCS as GWCS


@pytest.mark.parametrize(
    ('ext_list', 'n_dc'),
    [(None, 1),
     ('data', 1),
     (['data', 'var_rnoise'], 2)])
def test_roman_wfi_ext_options(imviz_helper, roman_imagemodel, ext_list, n_dc):
    imviz_helper.load_data(roman_imagemodel, data_label='roman_wfi_image_model', ext=ext_list)
    dc = imviz_helper.app.data_collection
    assert len(dc) == n_dc

    if ext_list == '*':
        ext_list = ('data', 'dq', 'err', 'var_poisson', 'var_rnoise')
    elif isinstance(ext_list, str):
        ext_list = (ext_list, )
    elif ext_list is None:
        ext_list = ('data', )

    for data, ext in zip(dc, ext_list):
        assert data.label == f'roman_wfi_image_model[{ext.upper()}]'
        assert data.shape == (20, 10)  # ny, nx
        assert isinstance(data.coords, GWCS)



import pytest
from astropy.io import fits

from jdaviz.utils import standardize_metadata, COMMENTCARD_KEY


def test_metadata_plain_dict():
    metadata = {'a': 1, 'b': 2, 'c': {'d': 42}}
    out_meta = standardize_metadata(metadata)
    assert out_meta == metadata

    # Make sure input is unchanged.
    del out_meta['c']
    assert out_meta == {'a': 1, 'b': 2}
    assert 'c' in metadata


def test_metadata_nested_fits_header():
    hdu = fits.PrimaryHDU()
    metadata = {'a': 1, 'header': hdu.header}
    out_meta = standardize_metadata(metadata)
    assert 'header' not in out_meta
    assert sorted(out_meta.keys()) == ['BITPIX', 'EXTEND', 'NAXIS', 'SIMPLE', COMMENTCARD_KEY, 'a']
    assert out_meta[COMMENTCARD_KEY]['BITPIX'] == 'array data type'
    assert 'BITPIX' in out_meta[COMMENTCARD_KEY]._header
    assert 'a' not in out_meta[COMMENTCARD_KEY]._header

    # Make sure input is unchanged.
    assert 'header' in metadata


def test_metadata_fits_header():
    hdu = fits.PrimaryHDU()
    out_meta = standardize_metadata(hdu.header)
    assert sorted(out_meta.keys()) == ['BITPIX', 'EXTEND', 'NAXIS', 'SIMPLE', COMMENTCARD_KEY]
    assert out_meta[COMMENTCARD_KEY]['BITPIX'] == 'array data type'
    assert 'BITPIX' in out_meta[COMMENTCARD_KEY]._header

    # Make sure input is unchanged.
    del out_meta['BITPIX']
    assert 'BITPIX' in hdu.header

    # Make sure you can still nest it afterwards if you want.
    # Similar logic is used to separate primary header in Metadata Viewer plugin.
    hdu_2 = fits.ImageHDU()
    out_meta['image_meta'] = standardize_metadata(hdu_2.header)
    assert out_meta['image_meta']['XTENSION'] == 'IMAGE'


def test_metadata_invalid():
    with pytest.raises(TypeError, match='metadata must be dictionary or FITS header'):
        standardize_metadata([1, 2, 3])



from specutils import Spectrum1D

from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection


__all__ = ['Spectrum2DImporter']


@loader_importer_registry('2D Spectrum')
class Spectrum2DImporter(BaseImporterToDataCollection):
    template_file = __file__, "../to_dc_with_label.vue"

    @property
    def is_valid(self):
        if self.app.config not in ('specviz', 'specviz2d'):
            # NOTE: temporary during deconfig process
            return False
        return isinstance(self.input, Spectrum1D) and self.input.flux.ndim == 2

    @property
    def default_viewer(self):
        # returns the registry name of the default viewer
        # only used if `show_in_viewer=True` and no existing viewers can accept the data
        return 'mosviz-profile-2d-viewer'
